<?php
/**
* Blesta <-> Omnipay glue code
*
* @package blesta
* @subpackage blesta.components.gateways
* @copyright Copyright (c) 2014 Floris Bos (TA Maxnet)
* @license Public domain
*/

/* Use phpseclib to generate random session ids */
require_once(VENDORDIR."phpseclib".DS."Crypt".DS."Random.php");

/* Use Record for storing sessions in database */
require_once(COMPONENTDIR."record".DS."record.php");

/* Offload Omnipay class loading to Composer */
require_once(VENDORDIR."autoload.php");


use Omnipay\Omnipay;
use Omnipay\Common\CreditCard;

abstract class OmnipayNonmerchantGateway extends NonmerchantGateway {

    protected $gw, $currency, $meta, $callbackurl;

    /* Using instance variables instead of static constants so derived classes can override */    
    protected $gwname = "";
    protected $friendlyname = "";
    protected $version = "1.0.0";
    protected $authors = array(
        array("name" => "Maxnet", "url" => "http://www.noc-ps.com/"),
        array("name" => "using Omnipay library", "url" => "https://github.com/omnipay/omnipay"));
    protected $currencies = array(
        "AED", "AFN", "ALL", "AMD", "ANG", "AOA", "ARS", "AUD",
        "AWG", "AZN", "BAM", "BBD", "BDT", "BGN", "BHD", "BIF", "BMD", "BND",
        "BOB", "BRL", "BSD", "BTC", "BTN", "BWP", "BYR", "BZD", "CAD", "CDF",
        "CHF", "CLP", "CNY", "COP", "CRC", "CUC", "CUP", "CVE", "CZK", "DJF",
        "DKK", "DOP", "DZD", "EGP", "ERN", "ETB", "EUR", "FJD", "FKP", "GBP",
        "GEL", "GGP", "GHS", "GIP", "GMD", "GNF", "GTQ", "GYD", "HKD", "HNL",
        "HRK", "HTG", "HUF", "IDR", "ILS", "IMP", "INR", "IQD", "IRR", "ISK",
        "JEP", "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRW",
        "KWD", "KYD", "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LVL",
        "LYD", "MAD", "MDL", "MGA", "MKD", "MMK", "MNT", "MOP", "MRO", "MUR",
        "MVR", "MWK", "MXN", "MYR", "MZN", "NAD", "NGN", "NIO", "NOK", "NPR",
        "NZD", "OMR", "PAB", "PEN", "PGK", "PHP", "PKR", "PLN", "PYG", "QAR",
        "RON", "RSD", "RUB", "RWF", "SAR", "SBD", "SCR", "SDG", "SEK", "SGD",
        "SHP", "SLL", "SOS", "SPL", "SRD", "STD", "SVC", "SYP", "SZL", "THB",
        "TJS", "TMT", "TND", "TOP", "TRY", "TTD", "TVD", "TWD", "TZS", "UAH",
        "UGX", "USD", "UYU", "UZS", "VEF", "VND", "VUV", "WST", "XAF", "XCD",
        "XDR", "XOF", "XPF", "YER", "ZAR", "ZMW", "ZWD");    

    function __construct()
    {
	Loader::loadComponents($this, array("Input"));
        
        if (!$this->gwname)
            throw new Exception("Derived class must specify \$gwname");
        if (!$this->friendlyname)
            $this->friendlyname = $this->gwname;
        if (!$this->callbackurl)
            $this->callbackurl = Configure::get("Blesta.gw_callback_url").Configure::get("Blesta.company_id")."/".strtolower(get_class($this))."/";
        
        $this->gw = Omnipay::create($this->gwname);
        if (!$this->gw->supportsCompletePurchase() )
            throw new Exception("We only support non-merchant gateways for now");
    }
    
    function getName()
    {
        return $this->friendlyname;
    }
    
    function getVersion()
    {
        return $this->version;
    }
    
    function getLogo()
    {
        return "logo.png";
    }
    
    function getAuthors()
    {
        return $this->authors;
    }
    
    function getCurrencies()
    {
        return $this->currencies;
    }
    
    function setCurrency($currency)
    {
        $this->currency = $currency;
    }
    
    function setMeta(array $meta=null)
    {
        $this->meta = $meta;
    }
    
    protected function logError($msg)
    {
        $this->log($_SERVER['REQUEST_URI'], $msg, 'input', false);
    }
    
    function install()
    {
        $r = new Record();
        $r->setField("id", array('type' => "varchar", 'size' => 255))
          ->setField("gateway", array('type' => "varchar", 'size' => 255))
          ->setField("expire", array('type' => "datetime"))
          ->setField("value", array('type' => "text", 'is_null' => true))
          ->setKey(array("id","gateway"), "primary")
	  ->create("gateway_sessions", true);    
    }
    
    function uninstall($id, $last_instance)
    {
        /* Probably should delete db tables here, but there may be more than one module derived from this class using them */
    }
    
    protected function getSession($id)
    {
        $r = new Record();
        $row = $r->select("value")->from("gateway_sessions")
            ->where("id", "=", $id)
            ->where("gateway", "=", $this->gwname)->fetch(PDO::FETCH_ASSOC);

        if ($row)
            return unserialize($row["value"]);
        else
            return false;
    }
    
    protected function putSession($id, $data)
    {
        $r = new Record();
        $r->where("id", "=", $id)
          ->where("gateway", "=", $this->gwname)
          ->update("gateway_sessions", array("value" => serialize($data)));
    }
    
    protected function createSessionID()
    {
        $r = new Record();
        $expires = date("Y-m-d H:i:s", time() + 86400 * 14);
        
        for ($try = 0; $try < 10; $try++)
        {
            try
            {
                $key = $this->_generateSessionKey();
                $r->insert("gateway_sessions", array('id' => $key, 'gateway' => $this->gwname, 'expire' => $expires));
                return $key;
            }
            catch (PDOException $e) { }
        }
        
        throw new Exception("Error creating session");
    }
    
    protected function _generateSessionKey()
    {
        return dechex(time()).dechex(crypt_random()).dechex(crypt_random());
    }
    
    /**
     * Create and return the view content required to modify the settings of this gateway
     *
     * @param array $meta An array of meta (settings) data belonging to this gateway
     * @return string HTML content containing the fields to update the meta data for this gateway
     */
    function getSettings(array $meta=null)
    {
        /* Dynamically generate form based on gateway's default parameters */
        Loader::loadHelpers($this, array("Form", "Html"));
        $defaults = $this->gw->getDefaultParameters();
        $html = "<ul>";
        $this->Form->setOutput(true);
        
        foreach ($defaults as $key => $defaultValue)
        {
            $html .= "<li>";
            $html .= $this->Form->label($key, "_".$key."_id");
            if (is_array($defaultValue))
            {
                $options = array();
                foreach ($defaultValue as $option)
                {
                    $options[$option] = $option;
                }
                $html .= $this->Form->fieldSelect($key, $options, $this->ifSet($meta[$key], reset($defaultValue)), array('id'=> "_".$key."_id"));
            }
            else if (is_bool($defaultValue))
            {
                $defaultValue = $defaultValue ? "1" : "0";
                $html .= $this->Form->fieldSelect($key, array("1" => "true", "0" => "false"), $this->ifSet($meta[$key], $defaultValue), array('id'=> "_".$key."_id"));
            }
            else
            {
                $html .= $this->Form->fieldText($key, $this->ifSet($meta[$key], $defaultValue), array('id'=> "_".$key."_id"));
            }
        }
        
        $html .= "</ul>";
        return $html;
    }
    
    /**
     * Validates the given meta (settings) data to be updated for this gateway
     *
     * @param array $meta An array of meta (settings) data to be updated for this gateway
     * @return array The meta data to be updated in the database for this gateway, or reset into the form on failure
     */
    function editSettings(array $meta)
    {
        return $meta;
    }
    
    /**
     * Returns an array of all fields to encrypt when storing in the database
     *
     * @return array An array of the field names to encrypt when storing in the database
     */
    function encryptableFields()
    {
        /* Just encrypt ALL fields */
        return array_keys($this->gw->getDefaultParameters());
    }

    /**
     * Returns all HTML markup required to render an authorization and capture payment form
     *
     * @param array $contact_info An array of contact info including:
     * 	- id The contact ID
     * 	- client_id The ID of the client this contact belongs to
     * 	- user_id The user ID this contact belongs to (if any)
     * 	- contact_type The type of contact
     * 	- contact_type_id The ID of the contact type
     * 	- first_name The first name on the contact
     * 	- last_name The last name on the contact
     * 	- title The title of the contact
     * 	- company The company name of the contact
     * 	- address1 The address 1 line of the contact
     * 	- address2 The address 2 line of the contact
     * 	- city The city of the contact
     * 	- state An array of state info including:
     * 		- code The 2 or 3-character state code
     * 		- name The local name of the country
     * 	- country An array of country info including:
     * 		- alpha2 The 2-character country code
     * 		- alpha3 The 3-cahracter country code
     * 		- name The english name of the country
     * 		- alt_name The local name of the country
     * 	- zip The zip/postal code of the contact
     * @param float $amount The amount to charge this contact
     * @param array $invoice_amounts An array of invoices, each containing:
     * 	- id The ID of the invoice being processed
     * 	- amount The amount being processed for this invoice (which is included in $amount)
     * @param array $options An array of options including:
     * 	- description The Description of the charge
     * 	- recur An array of recurring info including:
     * 		- amount The amount to recur
     * 		- term The term to recur
     * 		- period The recurring period (day, week, month, year, onetime) used in conjunction with term in order to determine the next recurring payment
     * @return mixed A string of HTML markup required to render an authorization and capture payment form, or an array of HTML markup
     */
    function buildProcess(array $contact_info, $amount, array $invoice_amounts=null, array $options=null)
    {
        try
        {
            $sessionID = $this->createSessionID();
            
            /* Why doesn't Blesta supply us with e-mail address and phone by default? */
            if (!isset($contact_info['email']) || !isset($contact_info['phone']))
            {
                Loader::loadModels($this, array("Contacts"));
                $contact = $this->Contacts->get($contact_info["id"]);
                $contact_info["email"] = $contact->email;
                $phone = '';
                $numbers = $this->Contacts->getNumbers($contact_info["id"], "phone");
                if (count($numbers))
                {
                    $phone = reset($numbers)->number;
                }
                
                $contact_info["phone"] = $phone;
            }
            
            $info = array(
                "firstName" => $contact_info['first_name'],
                "lastName"  => $contact_info['last_name'],
                "company"   => $contact_info['company'],
                "email"     => $contact_info['email'],
                "billingAddress1" => $contact_info['address1'],
                "billingAddress2" => $contact_info['address2'],
                "billingCity"     => $contact_info['city'],
                "billingPostcode" => $contact_info['zip'],
                "billingCountry"  => $contact_info['country']['alpha2'],
                "billingPhone"    => $contact_info['phone'],
                "billingState"    => $contact_info['state']['code'] /* Should this be included for non-US non-Canada addresses? */
            );
            
            /* Note that despite being called CreditCard we only pass the address information and let the gateway collect card information */
            $card = new CreditCard($info);

            $request = array(
                "card" => $card,
                "amount" => $amount,
                "currency" => $this->currency,
                "description" => $this->ifSet($options['description']),
                "transactionId" => $sessionID,
                "clientIp" => $_SERVER["REMOTE_ADDR"],
                "returnUrl" => $this->callbackurl.$sessionID,
                "cancelUrl" => ($_SERVER["SERVER_PORT"] == 443 ? "https://" : "http://").$_SERVER["HTTP_HOST"].$_SERVER["REQUEST_URI"]
            );
    
            $this->gw->initialize($this->meta);
            $response = $this->gw->purchase($request)->send();
            
            if ($response->isSuccessful())
            {
                throw new Exception("Unexpected payment success, non-merchant gateway should redirect first");
            }
            else if ($response->isRedirect())
            {
                $request['transactionReference'] = $response->getTransactionReference();
                $request['client_id'] = $contact_info['client_id'];
                $request['invoice_amounts'] = $invoice_amounts;
                unset($request['card']);
                $this->putSession($sessionID, $request);
                
                /* Workaround because Blesta's redirect magic does not preserve the session ID we appended to the callback URL
                   when it redirects the client to /client/pay/received/gateway at app/controllers/callback.php line 85 */
                $_SESSION["payment_session_id"] = $sessionID;

                $response->redirect();
                exit(0);
            }
            else
            {
                /* Error condition that should be displayed to user */
                $this->logError($response->getMessage());
                /* FIXME: what is the proper way of error reporting here? Input->setErrors() does not seem to work */
                return "<b>Error: ".htmlentities($response->getMessage())."</b>";
            }
        }
        catch (Exception $e)
        {
            /* Generic error conditions that should not be displayed to user */
            $this->logError( $e->getMessage() );
            /* FIXME: what is the proper way of error reporting here? Input->setErrors() does not seem to work */
            return "<b>Internal error. See logs for more information.</b>";
        }
    }
    
    /**
     * Validates the incoming POST/GET response from the gateway to ensure it is
     * legitimate and can be trusted.
     *
     * @param array $get The GET data for this request
     * @param array $post The POST data for this request
     * @return array An array of transaction data, sets any errors using Input if the data fails to validate
     *  - client_id The ID of the client that attempted the payment
     *  - amount The amount of the payment
     *  - currency The currency of the payment
     *  - invoices An array of invoices and the amount the payment should be applied to (if any) including:
     *  	- id The ID of the invoice to apply to
     *  	- amount The amount to apply to the invoice
     * 	- status The status of the transaction (approved, declined, void, pending, reconciled, refunded, returned)
     * 	- reference_id The reference ID for gateway-only use with this transaction (optional)
     * 	- transaction_id The ID returned by the gateway to identify this transaction
     * 	- parent_transaction_id The ID returned by the gateway to identify this transaction's original transaction (in the case of refunds)
     */
    function validate(array $get, array $post)
    {
        try
        {
            if (empty($get[2]))
                throw new Exception("Missing transaction ID in URL");

            $sessionKey = $get[2];
            
            $request = $this->getSession($sessionKey);
            if (!$request)
                throw new Exception("Unknown transaction ID");

            $this->gw->initialize($this->meta);
            $response = $this->gw->completePurchase($request)->send();

            if ($response->isSuccessful())
            {
                return array(
                    "client_id" => $request['client_id'],
                    "amount"    => $request['amount'],
                    "currency"  => $request['currency'],
                    "invoices"  => $request['invoice_amounts'],
                    "status"    => "approved",
                    "transaction_id" => $this->ifSet($response->getTransactionReference(), $sessionKey)
                );
            }
            else if ( (method_exists($response, "isInitialized") && $response->isInitialized() )
                     || (method_exists($response, "isUncleared") && $response->isUncleared() ))
            {
                /* Offline payments and payments that need to be manually approved due to high fraud risk */
                return array(
                    "client_id" => $request['client_id'],
                    "amount"    => $request['amount'],
                    "currency"  => $request['currency'],
                    "invoices"  => $request['invoice_amounts'],
                    "status"    => "pending",
                    "transaction_id" => $this->ifSet($response->getTransactionReference(), $sessionKey)
                );
            }
            else
            {
                /* Error condition that should be displayed to user */
                $this->logError($response->getMessage());
                $this->Input->setErrors(array("" => array("error" => $response->getMessage())));
            }
        }
        catch (Exception $e)
        {
            /* Generic error conditions that should not be displayed to user */
            $this->logError( $e->getMessage() );
            $this->Input->setErrors($this->getCommonError("general"));            
        }
    }
    
    /**
     * Returns data regarding a success transaction. This method is invoked when
     * a client returns from the non-merchant gateway's web site back to Blesta.
     *
     * @param array $get The GET data for this request
     * @param array $post The POST data for this request
     * @return array An array of transaction data, may set errors using Input if the data appears invalid
     *  - client_id The ID of the client that attempted the payment
     *  - amount The amount of the payment
     *  - currency The currency of the payment
     *  - invoices An array of invoices and the amount the payment should be applied to (if any) including:
     *  	- id The ID of the invoice to apply to
     *  	- amount The amount to apply to the invoice
     * 	- status The status of the transaction (approved, declined, void, pending, reconciled, refunded, returned)
     * 	- transaction_id The ID returned by the gateway to identify this transaction
     * 	- parent_transaction_id The ID returned by the gateway to identify this transaction's original transaction
     */
    function success(array $get, array $post)
    {
        /* Workaround because Blesta's redirect magic does not preserve the session ID we appended to the callback URL
           when it redirects the client to /client/pay/received/gateway at app/controllers/callback.php line 85 */
        if (empty($get[2]))
            $get[2] = $_SESSION["payment_session_id"];
        
        /* FIXME: Use a cached response instead? */
        return $this->validate($get, $post);
    }
}
